/*
 * Copyright (C) 2017 OpenSIPS Project
 *
 * This file is part of opensips, a free SIP server.
 *
 * opensips is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * opensips is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
#ifndef _SIPMSGOPS_LIST_HDR_H
#define _SIPMSGOPS_LIST_HDR_H

#include "../../dprint.h"
#include "../../ut.h"


int list_hdr_has_val(struct sip_msg *msg, int_str_t *match_hdr, str *val);

int list_hdr_add_val(struct sip_msg *msg, int_str_t *match_hdr, str *val);

int list_hdr_remove_val(struct sip_msg *msg, int_str_t *match_hdr, str *val);

#endif
